#![allow(dead_code, unused)]
use core::panic;
use std::ops::Sub;
use crate::client::*;
use crate::models::*;
use serde::{Serialize, Deserialize};

mod models;
mod client;

#[derive(PartialEq, Eq, Debug)]
enum Mode {
    Unset,
    Errornous,
    Zone,
    Record,
    PrimaryServer,
}

#[derive(PartialEq, Eq, Debug)]
enum SubMode {
    Help,
    Get,
    Create,
    Update,
    Delete,
}

#[derive(Debug)]
struct ZoneContext {
    all: bool,
    zone: String,
    name: String,
    ttl: u64,
}

#[derive(Debug)]
struct RecordProto<'a> {
    r#type: RecordType,
    name: &'a str,
    value: &'a str,
    ttl: Option<u64>,
}

#[derive(Debug)]
struct RecordContext<'a> {
    all: bool,
    zone: &'a str,
    records: Vec<RecordProto<'a>>,
}

#[derive(Debug)]
struct Context<'a> {
    mode: Mode,
    submode: SubMode,
    token: Option<String>,
    zone_context: ZoneContext,
    record_context: RecordContext<'a>,
}

#[tokio::main(flavor = "current_thread")]
async fn main() {
    let mut ctx = Context {
        mode: Mode::Unset,
        submode: SubMode::Help,
        token: std::env::var("HETZNER_DNS_API_TOKEN").ok(),
        zone_context: ZoneContext {
            all: false,
            zone: String::new(),
            name: String::new(),
            ttl: 86400,
        },
        record_context: RecordContext {
            all: false,
            zone: "",
            records: vec![],
        },
    };
    let mut _continue = false;
    for (idx, arg) in std::env::args().enumerate() {
        if _continue {
            continue;
        }
        match idx {
            0 => continue,
            1 => {
                ctx.mode = match arg.as_str() {
                    "zones" => Mode::Zone,
                    "records" => Mode::Record,
                    "primary" => Mode::PrimaryServer,
                    "z" => Mode::Zone,
                    "r" => Mode::Record,
                    "p" => Mode::PrimaryServer,
                    _ => Mode::Errornous,
                }
            }
            2 => {
                ctx.submode = match arg.as_str() {
                    "get" => SubMode::Get,
                    "create" => SubMode::Create,
                    "update" => SubMode::Update,
                    "delete" => SubMode::Delete,
                    "g" => SubMode::Get,
                    "c" => SubMode::Create,
                    "u" => SubMode::Update,
                    "d" => SubMode::Delete,
                    _ => SubMode::Help,
                }
            }
            _ => match ctx.mode {
                Mode::Zone => {
                    if arg.starts_with("-") {
                        match arg.as_str() {
                            "--all" => ctx.zone_context.all = true,
                            "--name" => {
                                ctx.zone_context.name = std::env::args().nth(idx + 1).unwrap()
                            }
                            "--ttl" => {
                                ctx.zone_context.ttl = std::env::args()
                                    .nth(idx + 1)
                                    .unwrap()
                                    .parse()
                                    .unwrap_or(86400)
                            }
                            "--zone" => {
                                ctx.zone_context.zone = std::env::args().nth(idx + 1).unwrap()
                            }
                            _ => panic!("unknown parameter {}", arg),
                        }
                    } else {
                        continue; // value, ignore
                    }
                }
                Mode::Record => {
                    if arg.starts_with("-") {
                        match arg.as_str() {
                            "--all" => ctx.record_context.all = true,
                            _ => todo!()
                        }
                    }
                }
                Mode::PrimaryServer => {
                    todo!()
                }
                _ => {
                    continue;
                }
            },
        }
    }

    if ctx.mode != Mode::Unset && ctx.mode != Mode::Errornous && ctx.submode != SubMode::Help {
        if let Some(ref token) = ctx.token {
            let client = HetznerDNSAPIClient::new(String::from(token.trim()));
            match ctx.mode {
                Mode::Zone => match ctx.submode {
                    SubMode::Get => {
                        if ctx.zone_context.all {
                            println!(
                                "{:#?}",
                                client.get_zones(None, None, None, None).await.unwrap()
                            );
                        }
                    }
                    _ => {
                        todo!()
                    }
                },
                Mode::Record => match ctx.submode {
                    SubMode::Get => {
                        if ctx.record_context.all {
                            println!("{:#?}", client.get_records(None, None, None).await.unwrap());
                        }
                    }
                    _ => { todo!() }
                }
                Mode::PrimaryServer => {
                    todo!()
                }
                _ => panic!("how in the even"),
            }
        } else {
            panic!("missing token!");
        }
    } else {
        match ctx.mode {
            Mode::Zone => println!("zone help"),
            Mode::Record => println!("record help"),
            Mode::PrimaryServer => println!("primary server help"),
            _ => println!("full help"),
        }
    }
}
